#
#  commoncode.py
#  ENRAGE
#

import random, string

commoncodeVersion = 3.7

def getMergedRegions(regionfilename, maxDist=1000, minHits=0, verbose=False, keepLabel=False, chromField=1, pad=0):
    """ returns a list of merged regions that are closer than maxDist 
    basepair; can optionally filter regions that have fewer than minHits.
    Can also optionally return the label of each region.
    """
    regions = {}
    infile = open(regionfilename)
    mergeCount = 0
    chromField = int(chromField)
    for line in infile:
        fields = line.split('\t')
        #print fields
        try:
            hits = float(fields[4].strip())
        except:
            continue
        if hits < minHits:
            continue
        if chromField > 1:
            label = string.join(fields[:chromField],'\t')
            chrom = fields[chromField][3:]
            start = int(fields[chromField + 1]) - pad
            stop = int(fields[chromField + 2]) + pad
        else:
            label = fields[0]
            chrom = fields[1][3:]
            start = int(fields[2]) - pad
            stop = int(fields[3]) + pad
        length = abs(stop - start)
        if chrom not in regions:
            regions[chrom] = []
        merged = False
        
        if len(regions[chrom]) > 0:
            for index in range(len(regions[chrom])):
                if keepLabel:
                    (rlabel, rstart, rstop, rlen) = regions[chrom][index]
                else:
                    (rstart, rstop, rlen) = regions[chrom][index]
                midpoint = (rstart + rstop) / 2
                #if abs(midpoint - start) < maxDist or abs(midpoint - stop) < maxDist or start == rstart or stop == rstop:
                if (rstart <= start <= rstop) or (rstart <= stop <= rstop) or (start <= rstart <= stop) or (start <= rstop <= stop):
                    if start < rstart:
                        rstart = start
                    if rstop < stop:
                        rstop = stop
                    rlen = abs(rstop - rstart)
                    if keepLabel:
                        regions[chrom][index] = (label, rstart, rstop, rlen)
                    else:
                        regions[chrom][index] = (rstart, rstop, rlen)
                    #print (rstart, rstop, rlen)
                    mergeCount += 1
                    merged = True
                    break
        
        if not merged:
            if keepLabel:
                regions[chrom].append((label, start, stop, length))
            else:
                regions[chrom].append((start, stop, length))
    infile.close()
    
    regionCount = 0
    for chrom in regions:
        regionCount += len(regions[chrom])
        if keepLabel:
            regions[chrom].sort(cmp=lambda x,y:cmp(x[2], y[2]))
        else:
            regions[chrom].sort()
    if verbose:
        print 'merged %d times' % mergeCount
        print 'returning %d regions' % regionCount
    
    return regions

def readSize(filename, verbose=False):
    """ returns the size of reads in bed file. Supports both regular
        reads and spliced reads.
    """
    hitfile = open(filename)
    hitfile.readline()
    line = hitfile.readline()
    hitfile.close()
    fields = line.strip().split()
    readlen = 0
    if len(fields) == 12:
        sizes = fields[10].split(',')
        for size in sizes:
            readlen += int(size)
    else:
        readlen = abs(int(fields[1]) - int(fields[2])) + 1
    if verbose:
        print 'readSize for %s: %d mer' % (filename, readlen)
    
    return readlen

def datasetSize(filename, verbose=False, countLabels=False):
    """ returns the size of a bed file. Supports both regular
    reads and multi-reads with the countLabels option.
    """
    count = 0
    prevLabel = ''
    bedfile = open(filename)
    for line in bedfile:
        if 'track' in line:
            continue
        if countLabels:
            fields = line.strip().split()
            if fields[3] != prevLabel:
                count += 1
                prevLabel = fields[3]
        else:
            count += 1
    if verbose:
        print '%s: %d reads' % (filename, count)
    
    return count

def getReadDict(filename, maxReads=0, filterReads=[], verbose=False, bothEnds=False, noSense=False, fullChrom=False):
    """ returns a dictionary of reads by chromosome from a bed 
    file. Can optionally return only maxReads number of reads, 
    filter reads that have a particular identifier (e.g. chromosome). 
    Normally returns only (start, sense) of reads, but can also 
    include (stop, sense) if bothEnds is equal to True.
    """
    index = 0
    hitDict = {}
    lines = open(filename)
    if maxReads > 0:
        hitfile = open(filename)
        lines = hitfile.readlines()
        hitfile.close()
        random.shuffle(lines)
        lines = lines[:maxReads]
    
    for line in lines:
        if 'track' in line:
            continue
        doContinue = False
        for stem in filterReads:
            if stem in line:
                doContinue = True
                break
        if doContinue:
            continue
        fields = line.split()
        if fullChrom:
            chrom = fields[0]
        else:
            chrom = fields[0][3:]
        if chrom not in hitDict:
            hitDict[chrom] = []
        if bothEnds:
            if noSense:
                hitDict[chrom].append((int(fields[1]), int(fields[2])))
            else:
                hitDict[chrom].append((int(fields[1]), int(fields[2]), fields[5]))
        else:
            if noSense:
                hitDict[chrom].append(int(fields[1]))
            else:
                hitDict[chrom].append((int(fields[1]), fields[5]))
        index += 1
        if verbose and index % 1000000 == 0:
            print 'read %d' % index
    try:
        lines.close()
    except:
        lines = []
    
    for chrom in hitDict:
        hitDict[chrom].sort()
    if verbose:
        print 'returning %d reads' % index
    
    return hitDict

def readDictComplexity(hitDict):
    """ returns the count of unique reads and all reads to get 
    a feel for the complexity of the library. 
    """
    uniqueCount = 0
    totalCount = 0
    currentRead = ''
    for chrom in hitDict:
        for read in hitDict[chrom]:
            if read != currentRead:
                uniqueCount += 1
                currentRead = read
            totalCount += 1
    return (uniqueCount, totalCount)
    
def getMultiReadDict(filename, verbose=False):
    """ returns a dictionary of multiple-match reads by chromosome, 
    with reads as (start, readID).
    Can also be used on unique reads (labeled as 'read'), for unique
    read weighing.
    """
    index = 0
    hitfile = open(filename)
    hitfile.readline()
    hitDict = {}
    
    for line in hitfile:
        fields = line.split()
        try:
            if 'read' in fields[3]:
                readID = index
                readTimes = '-1'
            else:
                (readID,readTimes) = fields[3].split('x')
        except:
            print 'could not process line: %s' % line.strip()
            continue
        if readTimes == '1':
            continue
        chrom = fields[0][3:]
        if chrom not in hitDict:
            hitDict[chrom] = []
        hitDict[chrom].append((int(fields[1]), int(readID)))
        index += 1
    hitfile.close()
    
    for chrom in hitDict:
        hitDict[chrom].sort()
    
    if verbose:
        print 'returning %d reads' % index
    
    return hitDict

def getMultiReadIDDict(filename, verbose=False):
    """ returns a dictionary of multiple-match reads by readID, 
    with reads as (chrom, start).
    """
    index = 0
    hitfile = open(filename)
    hitfile.readline()
    hitDict = {}
    currentReadID = ''
    for line in hitfile:
        fields = line.split()
        (readID,readTimes) = fields[3].split('x')
        if readTimes == '1':
            continue
        readID = int(readID)
        chrom = fields[0][3:]
        if readID != currentReadID:
            hitDict[readID] = []
            currentReadID = readID
            index += 1
        if (chrom, int(fields[1])) not in hitDict[readID]:
            hitDict[readID].append((chrom, int(fields[1])))
    hitfile.close()
    if verbose:
        print 'returning %d readIDs' % index
    
    return hitDict

def findPeak(hitList, start, length, readlen=25):
    """ find the peak in a list of reads (hitlist) in a region 
    of a given length and absolute start point. returns a 
    list of peaks, the number of hits, a triangular-smoothed 
    version of hitlist, and the number of reads that are 
    forward (plus) sense.
    """
	
    seqArray = [0] * length
    smoothArray = [0] * length
    numHits = 0
    numPlus = 0
    for (hstart, sense) in hitList:
        currentpos = hstart - start
        if (currentpos <  1 - readlen) or (currentpos >= length):
            continue
        hitIndex = 0
        numHits += 1
        while currentpos < 0:
            hitIndex += 1
            currentpos += 1
        while hitIndex < readlen and  currentpos < length:
            seqArray[currentpos] += 1
            hitIndex += 1
            currentpos += 1
        if sense == '+':
            numPlus += 1
    
    # implementing a triangular smooth
    for pos in range(2,length -2):
        smoothArray[pos] = (seqArray[pos -2] + 2 * seqArray[pos - 1] + 3 * seqArray[pos] + 2 * seqArray[pos + 1] + seqArray[pos - 2]) / 9.0
    topNucleotide = 0
    topPos = []
    for currentpos in range(length):
        if topNucleotide < smoothArray[currentpos]:
            topNucleotide = smoothArray[currentpos]
            topPos = [currentpos]
        elif topNucleotide  == smoothArray[currentpos]:
            topPos.append(currentpos)
    
    return (topPos, numHits, smoothArray, numPlus)

def getFeaturesByChromDict(genomeObject, additionalRegionsDict={}, ignorePseudo=False, restrictList=[]):
    """ return a dictionary of cistematic gene features. Requires
    cistematic, obviously. Can filter-out pseudogenes. Will use
    additional regions dict to supplement gene models, if available. 
    Can restrict output to a list of GIDs.
    """ 
    featuresDict = genomeObject.getallGeneFeatures()
    restrictGID = False
    if len(restrictList) > 0:
        restrictGID = True
    if len(additionalRegionsDict) > 0:
        sortList = []
        for chrom in additionalRegionsDict:
            for (label, start, stop, length) in additionalRegionsDict[chrom]:
                if label not in sortList:
                    sortList.append(label)
                if label not in featuresDict:
                    featuresDict[label] = []
                    sense = '+'
                else:
                    sense = featuresDict[label][0][-1]
                featuresDict[label].append(('custom', chrom, start, stop, sense))
        for gid in sortList:
            featuresDict[gid].sort(cmp=lambda x,y:cmp(x[2], y[2]))

    featuresByChromDict = {}
    for gid in featuresDict:
        if restrictGID and gid not in restrictList:
            continue
        featureList = featuresDict[gid]
        newFeatureList = []
        isPseudo = False
        for (ftype, chrom, start, stop, sense) in featureList:
            if ftype == 'PSEUDO':
                isPseudo = True
            if (start, stop, ftype) not in newFeatureList:
                notContained = True
                containedList = []
                for (fstart, fstop, ftype2) in newFeatureList:
                    if start >= fstart and stop <= fstop:
                        notContained = False
                    if start < fstart and stop > fstop:
                        containedList.append((fstart, fstop))
                if len(containedList) > 0:
                    newFList = []
                    notContained = True
                    for (fstart, fstop, ftype2) in newFeatureList:
                        if (fstart, fstop) not in containedList:
                            newFList.append((fstart, fstop, ftype2))
                            if start >= fstart and stop <= fstop:
                                notContained = False
                    newFeatureList = newFList
                if notContained:
                    newFeatureList.append((start, stop, ftype))
        
        if ignorePseudo and isPseudo:
            continue
        if chrom not in featuresByChromDict:
            featuresByChromDict[chrom] = []
        for (start, stop, ftype) in newFeatureList:
            featuresByChromDict[chrom].append((start, stop, gid, sense, ftype))
	
    for chrom in featuresByChromDict:
		featuresByChromDict[chrom].sort()
	
    return featuresByChromDict

def getLocusByChromDict(genomeObject, upstream = 0, downstream = 0, useCDS=True, additionalRegionsDict={}, ignorePseudo=False, upstreamSpanTSS=False, keepSense=False, adjustToNeighbor=True):
    """ return a dictionary of gene loci. Can be used to retrieve additional 
    sequence upstream or downstream of gene, up to the next gene. Requires
    cistematic, obviously. 
    Can filter-out pseudogenes and use additional regions outside of existing 
    gene models. Use upstreamSpanTSS to overlap half of the upstream region 
    over the TSS.
    """ 
    locusByChromDict = {}
    if upstream == 0 and downstream == 0 and not useCDS:
        print 'getLocusByChromDict: asked for no sequence - returning empty dict'
        return locusByChromDict
    elif upstream > 0 and downstream > 0 and not useCDS:
        print 'getLocusByChromDict: asked for only upstream and downstream - returning empty dict'
        return locusByChromDict
    
    genome = genomeObject.genome
    featuresDict = genomeObject.getallGeneFeatures()
    if len(additionalRegionsDict) > 0:
        sortList = []
        for chrom in additionalRegionsDict:
            for (label, start, stop, length) in additionalRegionsDict[chrom]:
                if label not in sortList:
                    sortList.append(label)
                if label not in featuresDict:
                    featuresDict[label] = []
                    sense = '+'
                else:
                    sense = featuresDict[label][0][-1]
                featuresDict[label].append(('custom', chrom, start, stop, sense))
        for gid in sortList:
            featuresDict[gid].sort(cmp=lambda x,y:cmp(x[2], y[2]))
    
    for gid in featuresDict:
        featureList = featuresDict[gid]
        newFeatureList = []
        isPseudo = False
        for (ftype, chrom, start, stop, sense) in featureList:
            if ftype == 'PSEUDO':
                isPseudo = True
            newFeatureList.append((start, stop))
        if ignorePseudo and ftype == 'PSEUDO':
            continue
        newFeatureList.sort()
        
        sense = featureList[0][-1]
        gstart = newFeatureList[0][0]
        gstop = newFeatureList[-1][1]
        if sense == 'F':
            if not useCDS and upstream > 0:
                if upstreamSpanTSS:
                    if gstop > (gstart + upstream / 2):
                        gstop = gstart + upstream / 2
                    # otherwise leave gstop as is
                else:
                    gstop = gstart
            elif not useCDS and downstream > 0:
                gstart = gstop
            if upstream > 0:
                if upstreamSpanTSS:
                    distance = upstream / 2
                else:
                    distance = upstream
                if adjustToNeighbor:
                    nextGene = genomeObject.leftGeneDistance((genome, gid), distance * 2)
                    if nextGene < distance * 2:
                        distance = nextGene / 2
                if distance < 1:
                    distance = 1
                gstart -= distance
            if downstream > 0:
                distance = downstream
                if adjustToNeighbor:
                    nextGene = genomeObject.rightGeneDistance((genome, gid), downstream * 2)
                    if nextGene < downstream * 2:
                        distance = nextGene / 2
                if distance < 1:
                    distance = 1
                gstop += distance
        else:
            if not useCDS and upstream > 0:
                if upstreamSpanTSS:
                    if gstart < (gstop - upstream / 2):
                        gstart = gstop - upstream / 2
                    # otherwise leave gstart as is
                else:
                    gstart = gstop
            elif not useCDS and downstream > 0:
                    gstop = gstart
            if upstream > 0:
                if upstreamSpanTSS:
                    distance = upstream /2
                else:
                    distance = upstream
                if adjustToNeighbor:
                    nextGene = genomeObject.rightGeneDistance((genome, gid), distance * 2)
                    if nextGene < distance * 2:
                        distance = nextGene / 2
                if distance < 1:
                    distance = 1
                gstop += distance
            if downstream > 0:
                distance = downstream
                if adjustToNeighbor:
                    nextGene = genomeObject.leftGeneDistance((genome, gid), downstream * 2)
                    if nextGene < downstream * 2:
                        distance = nextGene / 2
                if distance < 1:
                    distance = 1
                gstart -= distance
        
        glen = abs(gstop - gstart)
        if chrom not in locusByChromDict:
            locusByChromDict[chrom] = []
        if keepSense:
            locusByChromDict[chrom].append((gstart, gstop, gid, glen, sense))
        else:
            locusByChromDict[chrom].append((gstart, gstop, gid, glen))
    
    for chrom in locusByChromDict:
        locusByChromDict[chrom].sort()
	
    return locusByChromDict

def computeRegionBins(regionsByChromDict, hitDict, bins, readlen, regionList = {}, normalizedTag=1., defaultRegionFormat = True, fixedFirstBin=-1, binLength=-1):
    """ returns 2 dictionaries of bin counts and region lengths, given a dictionary of predefined regions,
        a dictionary of reads, a number of bins, the length of reads, and optionally a list of regions 
        or a different weight / tag.
    """
    index = 0
    regionsBins = {}
    regionsLen = {}
    
    if defaultRegionFormat:
        regionIDField = 0
        startField = 1 
        stopField = 2
        lengthField = 3
    else:
        startField = 0
        stopField = 1
        regionIDField = 2
        lengthField = 3
    senseField = 4
    
    print 'entering computeRegionBins'
    if len(regionList) > 0:
        for readID in regionList:
            regionsBins[readID] = [0.] * bins
    else:
        for chrom in regionsByChromDict:
            for regionTuple in regionsByChromDict[chrom]:
                regionsBins[regionIDField] = [0.] * bins
    for chrom in hitDict:
        if chrom not in regionsByChromDict:
            continue
        for regionTuple in regionsByChromDict[chrom]:
            regionsLen[regionTuple[regionIDField]] = regionTuple[lengthField]
        print '\n' + chrom
        startRegion = 0
        for (tagStart, sense) in hitDict[chrom]:
            index += 1
            if index % 100000 == 0:
                print 'read %d ' % index,
            stopPoint = tagStart + readlen
            if startRegion < 0:
                startRegion = 0
            for regionTuple in regionsByChromDict[chrom][startRegion:]:
                start = regionTuple[startField]
                stop = regionTuple[stopField]
                regionID = regionTuple[regionIDField]
                rlen = regionTuple[lengthField]
                try:
                    rsense = regionTuple[senseField]
                except:
                    rsense = 'F'
                if tagStart > stop:
                    startRegion += 1
                    continue
                if start > stopPoint:
                    startRegion -= 10
                    break
                if start <= tagStart <= stop:
                    if binLength < 1:
                        regionBinLength = rlen / bins
                    else:
                        regionBinLength = binLength
                    startdist = tagStart - start
                    if rsense == 'F':
                        # we are relying on python's integer division quirk
                        binID = startdist / regionBinLength 
                        if (fixedFirstBin > 0) and (startdist < fixedFirstBin):
                            binID = 0
                        elif fixedFirstBin > 0:
                            binID = 1
                        if binID >= bins:
                            binID = bins - 1
                        try:
                            regionsBins[regionID][binID] += normalizedTag
                        except:
                            print '%s %s' % (regionID, str(binID))
                    else:
                        rdist = rlen - startdist
                        binID = rdist / regionBinLength 
                        if (fixedFirstBin > 0) and (rdist < fixedFirstBin):
                            binID = 0
                        elif fixedFirstBin > 0:
                            binID = 1
                        if binID >= bins:
                            binID = bins - 1
                        try:
                            regionsBins[regionID][binID] += normalizedTag
                        except:
                            print '%s %s' % (regionID, str(binID))
                    stopPoint = stop
    
    return (regionsBins, regionsLen)
